import re

from lunr.pipeline import Pipeline

full_re = re.compile(r"^\W*?([^\W]+)\W*?$")


def trimmer(token, i=None, tokens=None):
    def trim(s, metadata=None):
        match = full_re.match(s)
        if match is None:
            return s
        return match.group(1)

    return token.update(trim)


Pipeline.register_function(trimmer, "trimmer")



import json
import os
import re
import subprocess

import pytest

PATTERN = r'([^\ ]+) "([^\"]+)" \[([\d\.]*)\]'
DEFAULT_TOLERANCE = 1e-2


def assert_field_vectors_equal(a, b, tol=DEFAULT_TOLERANCE):
    assert a[0] == b[0]
    for x, y in zip(a[1], b[1]):
        assert x == pytest.approx(y, rel=tol)


def assert_vectors_equal(a, b, tol=DEFAULT_TOLERANCE):
    for x, y in zip(a, b):
        assert x == pytest.approx(y, rel=tol)


def assert_results_match(results, js_results, tol=DEFAULT_TOLERANCE):
    assert len(results) == len(js_results) != 0
    for js_result, result in zip(js_results, results):
        id_, title, score = re.match(PATTERN, js_result).groups()
        assert result["ref"] == id_
        assert result["score"] == pytest.approx(float(score), rel=tol)


def read_json_fixture(filename):
    fixture_path = os.path.join(
        os.path.dirname(__file__), "acceptance_tests", "fixtures", filename
    )
    with open(fixture_path) as f:
        return json.loads(f.read())


def run_node_script(filename, *args):
    js_path = os.path.join(
        os.path.dirname(__file__), "acceptance_tests", "javascript", filename
    )
    js_output = subprocess.check_output(["node", js_path] + list(args))
    return js_output.decode("utf-8").strip()



def get_language_stemmer(language):
    """Retrieves the SnowballStemmer for a particular language.

    Args:
        language (str): ISO-639-1 code of the language.
    """
    from lunr.languages import SUPPORTED_LANGUAGES
    from nltk.stem.snowball import SnowballStemmer  # type: ignore

    return SnowballStemmer(SUPPORTED_LANGUAGES[language])


def nltk_stemmer(stemmer, token, i=None, tokens=None):
    """Wrapper around a NLTK SnowballStemmer, which includes stop words for
    each language.

    Args:
        stemmer (SnowballStemmer): Stemmer instance that performs the stemming.
        token (lunr.Token): The token to stem.
        i (int): The index of the token in a set.
        tokens (list): A list of tokens representing the set.
    """

    def wrapped_stem(token, metadata=None):
        return stemmer.stem(token)

    return token.update(wrapped_stem)



import pytest

from lunr import lunr


@pytest.fixture
def documents():
    return [
        {
            "id": "a",
            "title": "Mr. Green kills Colonel Mustard",
            "body": """Mr. Green killed Colonel Mustard in the study with the
candlestick. Mr. Green is not a very nice fellow.""",
            "word_count": 19,
        },
        {
            "id": "b",
            "title": "Plumb waters plant",
            "body": "Professor Plumb has a green plant in his study",
            "word_count": 9,
        },
        {
            "id": "c",
            "title": "Scarlett helps Professor",
            "body": """Miss Scarlett watered Professor Plumbs green plant
while he was away from his office last week.""",
            "word_count": 16,
        },
    ]


@pytest.fixture
def index(documents):
    return lunr(ref="id", fields=("title", "body"), documents=documents)



from lunr.match_data import MatchData


class TestMatchData:
    def setup_method(self, method):
        self.match = MatchData("foo", "title", {"position": [1]})
        self.match.combine(MatchData("bar", "title", {"position": [2]}))
        self.match.combine(MatchData("baz", "body", {"position": [3]}))
        self.match.combine(MatchData("baz", "body", {"position": [4]}))

    def test_repr(self):
        assert repr(self.match) == '<MatchData "bar,baz,foo">'

    def test_create_empty_match_data(self):
        assert MatchData().metadata == {}

    def test_create_missing_field(self):
        assert MatchData("foo").metadata["foo"] == {}

    def test_create_missing_metadata(self):
        assert MatchData("foo", "title").metadata["foo"]["title"] == {}

    def test_combine_terms(self):
        assert sorted(list(self.match.metadata.keys())) == ["bar", "baz", "foo"]

    def test_combine_metadata(self):
        assert self.match.metadata["foo"]["title"]["position"] == [1]
        assert self.match.metadata["bar"]["title"]["position"] == [2]
        assert self.match.metadata["baz"]["body"]["position"] == [3, 4]

    def test_combine_does_not_mutate_source_data(self):
        metadata = {"foo": [1]}
        match_data1 = MatchData("foo", "title", metadata)
        match_data2 = MatchData("foo", "title", metadata)

        match_data1.combine(match_data2)

        assert metadata["foo"] == [1]

    def test_add_metadata_for_missing_term(self):
        self.match.add("spam", "title", {"position": [5]})

        assert self.match.metadata["spam"]["title"]["position"] == [5]

    def test_add_metadata_for_missing_field(self):
        self.match.add("foo", "body", {"position": [6]})

        assert self.match.metadata["foo"]["body"]["position"] == [6]

    def test_add_metadata_for_existing_term_field_and_metadata_key(self):
        self.match.add("foo", "title", {"position": [7]})

        assert self.match.metadata["foo"]["title"]["position"] == [1, 7]

    def test_add_metadata_for_existing_term_and_field_and_missing_metadata_key(self):
        self.match.add("foo", "title", {"weight": [7]})

        assert self.match.metadata["foo"]["title"] == {"position": [1], "weight": [7]}



from lunr import lunr, get_default_builder
from lunr.pipeline import Pipeline
from lunr.stemmer import stemmer
from lunr.trimmer import trimmer
from lunr.stop_word_filter import stop_word_filter

documents = [
    {
        "id": "a",
        "title": "Mr. Green kills Colonel Mustard",
        "body": """Mr. Green killed Colonel Mustard in the study with the
candlestick. Mr. Green is not a very nice fellow.""",
        "word_count": 19,
    },
    {
        "id": "b",
        "title": "Plumb waters plant",
        "body": "Professor Plumb has a green plant in his study",
        "word_count": 9,
    },
    {
        "id": "c",
        "title": "Scarlett helps Professor",
        "body": """Miss Scarlett watered Professor Plumbs green plant
while he was away from his office last week.""",
        "word_count": 16,
    },
]


def test_get_default_builder():
    builder = get_default_builder()
    assert builder.pipeline._stack == [trimmer, stop_word_filter, stemmer]
    assert builder.search_pipeline._stack == [stemmer]


def test_drop_pipeline_function():
    builder = get_default_builder()
    builder.pipeline.remove(stemmer)

    idx = lunr("id", ("title", "body"), documents, builder=builder)

    assert idx.search("kill") == []  # no match because "killed" was not stemmed


def test_add_token_metadata():
    builder = get_default_builder()

    def token_length(token, i, tokens):
        token.metadata["token_length"] = len(str(token))
        return token

    Pipeline.register_function(token_length)
    builder.pipeline.add(token_length)
    builder.metadata_whitelist.append("token_length")

    idx = lunr("id", ("title", "body"), documents, builder=builder)

    [result, _, _] = idx.search("green")
    assert result["match_data"].metadata["green"]["title"]["token_length"] == [5]
    assert result["match_data"].metadata["green"]["body"]["token_length"] == [5, 5]



import pytest

from lunr import lunr
from lunr.languages import LANGUAGE_SUPPORT, SUPPORTED_LANGUAGES
from lunr.pipeline import Pipeline

documents = [
    {
        "id": "a",
        "text": (
            "Este es un ejemplo inventado de lo que sería un documento en el "
            "idioma que se más se habla en España."
        ),
        "title": "Ejemplo de documento en español",
    },
    {
        "id": "b",
        "text": (
            "Según un estudio que me acabo de inventar porque soy un experto en"
            "idiomas que se hablan en España."
        ),
        "title": "Español es el tercer idioma más hablado del mundo",
    },
]


class TestLanguageSupport:
    @classmethod
    def setup_class(cls):
        assert (
            LANGUAGE_SUPPORT is True
        ), "NLTK not found, please run `pip install -e .[languages]`"

    def test_lunr_function_raises_if_unsupported_language(self):
        with pytest.raises(RuntimeError):
            lunr("id", ["title", "text"], documents, "foo")

    def test_lunr_function_raises_if_any_unsupported_language_is_passed(self):
        with pytest.raises(RuntimeError):
            lunr("id", ["title", "text"], documents, ["es", "foo"])

    def test_register_languages_in_pipeline_class(self):
        for lang in set(SUPPORTED_LANGUAGES) - {"en"}:
            assert "stemmer-{}".format(lang) in Pipeline.registered_functions

    def test_lunr_function_registers_nltk_stemmers_in_pipeline(self):
        idx = lunr("id", ["title", "text"], documents, ["es", "it"])
        assert "stemmer-es" in repr(idx.pipeline)
        assert "stemmer-it" in repr(idx.pipeline)

    def test_lunr_registers_lun_stemmers_in_pipeline_if_language_is_en(self):
        idx = lunr("id", ["title", "text"], documents, ["en", "es"])
        assert "stemmer,stemmer-es" in repr(idx.pipeline)

    def test_search_stems_search_terms(self):
        idx = lunr("id", ["title", "text"], documents, "es")
        results = idx.search("inventando")  # stemmed to "invent"
        assert len(results) == 2

    def test_search_stems_search_terms_for_both_languages(self):
        italian_document = {
            "id": "c",
            "text": (
                "Secondo uno studio che ho appena inventato perché sono un "
                "esperto di lingue parlate in Spagna."
            ),
            "title": "Lo spagnolo è la terza lingua più parlata al mondo",
        }
        idx = lunr(
            ref="id",
            fields=["title", "text"],
            documents=(documents + [italian_document]),
            languages=["es", "it"],
        )
        results = idx.search("spagna")
        assert len(results) == 1

        results = idx.search("inventando")
        assert len(results) == 2



import json
import tempfile

import pytest

from lunr import lunr
from lunr.index import Index
from tests.utils import read_json_fixture, run_node_script, assert_results_match


@pytest.mark.acceptance
def test_languages_query_results_match_javascript_results():
    query_string = "resistencia"
    js_results = run_node_script("language_query.js", query_string).split("\n")
    data = read_json_fixture("lang_es.json")
    index = lunr(
        ref="id", fields=("title", "text"), documents=data["docs"], languages="es"
    )
    results = index.search(query_string)
    assert_results_match(results, js_results, tol=0.1)


@pytest.mark.acceptance
def test_js_serialized_lang_index_can_be_loaded_and_produces_same_results():
    json_path = run_node_script("language_serialize_index.js")
    with open(json_path) as fd:
        js_serialized_index = fd.read()

    index = Index.load(js_serialized_index)
    query_string = "imperio"
    results = index.search(query_string)
    js_results = run_node_script("language_query.js", query_string).split("\n")
    assert_results_match(results, js_results)


@pytest.mark.acceptance
def test_serialized_lang_index_can_be_loaded_in_js_and_produces_same_results():
    data = read_json_fixture("lang_es.json")
    index = lunr(
        ref="id", fields=("title", "text"), documents=data["docs"], languages="es"
    )
    query_string = "imperio"
    results = index.search(query_string)
    serialized_index = index.serialize()

    with tempfile.NamedTemporaryFile(delete=False) as fp:
        fp.write(json.dumps(serialized_index).encode())

    js_results = run_node_script(
        "language_load_serialized_index_and_search.js", fp.name, query_string
    ).split("\n")
    assert_results_match(results, js_results)


@pytest.mark.acceptance
def test_serialized_multilang_index_can_be_loaded_in_js_and_results_equal():
    data = read_json_fixture("lang_es_en.json")
    index = lunr(
        ref="id",
        fields=("title", "text"),
        documents=data["docs"],
        languages=["es", "en"],
    )
    query_string = "taxation"
    results = index.search(query_string)
    serialized_index = index.serialize()

    with tempfile.NamedTemporaryFile(delete=False) as fp:
        fp.write(json.dumps(serialized_index).encode())

    js_results = run_node_script(
        "language_load_serialized_index_and_search.js",
        fp.name,
        query_string,
        "lang_es_en.json",
    ).split("\n")
    assert_results_match(results, js_results)



from math import sqrt

import pytest

from lunr.vector import Vector
from lunr.exceptions import BaseLunrException


def _vector_from_args(*args):
    vector = Vector()
    for i, arg in enumerate(args):
        vector.insert(i, arg)
    return vector


def test_vector_repr():
    vector = _vector_from_args(1, 3, -5)
    assert repr(vector) == "<Vector magnitude={}>".format(vector.magnitude)


class TestVectorPositionForIndex:
    vector = Vector([1, "a", 2, "b", 4, "c", 7, "d", 11, "e"])

    def test_position_for_index_at_the_beggining(self):
        assert self.vector.position_for_index(0) == 0

    def test_position_for_index_at_the_end(self):
        assert self.vector.position_for_index(20) == 10

    def test_position_for_index_consecutive(self):
        assert self.vector.position_for_index(3) == 4

    def test_position_for_index_non_consecutive_gap_after(self):
        assert self.vector.position_for_index(5) == 6

    def test_position_for_index_non_consecutive_gap_before(self):
        assert self.vector.position_for_index(6) == 6

    def test_position_for_index_non_consecutive_gap_before_and_after(self):
        assert self.vector.position_for_index(9) == 8

    def test_position_for_index_duplicate_at_the_beggining(self):
        assert self.vector.position_for_index(1) == 0

    def test_position_for_index_duplicate_at_the_end(self):
        assert self.vector.position_for_index(11) == 8

    def test_position_for_index_duplicate_consecutive(self):
        assert self.vector.position_for_index(4) == 4


def test_magnitude_calculates_magnitude():
    vector = _vector_from_args(4, 5, 6)
    assert sqrt(77) == vector.magnitude


def test_dot_calculates_dot_product_of_two_vectors():
    v1 = _vector_from_args(1, 3, -5)
    v2 = _vector_from_args(4, -2, -1)

    assert v1.dot(v2) == 3


class TestSimilarity:
    def test_similarity_calculates_the_similarity_between_two_vectors(self):
        v1 = _vector_from_args(1, 3, -5)
        v2 = _vector_from_args(4, -2, -1)

        assert v1.similarity(v2) == pytest.approx(0.5, 0.1)

    def test_empty_vector(self):
        v_empty = Vector()
        v1 = _vector_from_args(1)

        assert v1.similarity(v_empty) == 0
        assert v_empty.similarity(v1) == 0

    def test_non_overlapping_vector(self):
        v1 = Vector([1, 1])
        v2 = Vector([2, 1])

        assert v1.similarity(v2) == 0
        assert v2.similarity(v1) == 0


class TestVectorInsert:
    def test_insert_invalidates_magnitude_cache(self):
        vector = _vector_from_args(4, 5, 6)
        assert sqrt(77) == vector.magnitude

        vector.insert(3, 7)

        assert sqrt(126) == vector.magnitude

    def test_insert_keeps_items_in_index_specified_order(self):
        vector = Vector()

        vector.insert(2, 4)
        vector.insert(1, 5)
        vector.insert(0, 6)

        assert vector.to_list() == [6, 5, 4]

    def test_insert_fails_when_duplicate_entry(self):
        vector = _vector_from_args(4, 5, 6)
        with pytest.raises(BaseLunrException):
            vector.insert(0, 44)


class TestVectorUpsert:
    def test_upsert_invalidates_magnitude_cache(self):
        vector = _vector_from_args(4, 5, 6)
        assert vector.magnitude == sqrt(77)

        vector.upsert(3, 7)

        assert vector.magnitude == sqrt(126)

    def test_upsert_keeps_items_in_index_specified_order(self):
        vector = Vector()

        vector.upsert(2, 4)
        vector.upsert(1, 5)
        vector.upsert(0, 6)

        assert vector.to_list() == [6, 5, 4]

    def test_upsert_calls_fn_for_value_on_duplicate(self):
        vector = _vector_from_args(4, 5, 6)

        vector.upsert(0, 4, lambda current, passed: current + passed)

        assert vector.to_list() == [8, 5, 6]

    def test_upsert_defaults_to_passed_value_on_duplicate(self):
        vector = _vector_from_args(4, 5, 6)

        vector.upsert(0, 3)

        assert vector.to_list() == [3, 5, 6]



import pytest

from lunr.token_set import TokenSet
from lunr.exceptions import BaseLunrException


class TestTokenSetStr:
    def test_str_includes_node_finality(self):
        non_final = TokenSet()
        final = TokenSet()
        other_final = TokenSet()

        final.final = True
        other_final.final = True

        assert str(non_final) != str(final)
        assert str(other_final) == str(final)

    def test_str_includes_all_edges(self):
        zero_edges = TokenSet()
        one_edge = TokenSet()
        two_edges = TokenSet()

        one_edge.edges["a"] = 1
        two_edges.edges["a"] = 1
        two_edges.edges["b"] = 1

        assert str(zero_edges) != str(one_edge)
        assert str(two_edges) != str(one_edge)
        assert str(two_edges) != str(zero_edges)

    def test_str_includes_edge_id(self):
        child_a = TokenSet()
        child_b = TokenSet()
        parent_a = TokenSet()
        parent_b = TokenSet()
        parent_c = TokenSet()

        parent_a.edges["a"] = child_a
        parent_b.edges["a"] = child_b
        parent_c.edges["a"] = child_b

        assert str(parent_b) == str(parent_c)
        assert str(parent_a) != str(parent_c)
        assert str(parent_a) != str(parent_b)


class TestTokenSetFromString:
    def test_from_string_without_wildcard(self):
        TokenSet._next_id = 1
        x = TokenSet.from_string("a")

        assert str(x) == "0a2"
        assert x.edges["a"].final

    def test_from_string_with_trailing_wildcard(self):
        x = TokenSet.from_string("a*")
        wild = x.edges["a"].edges["*"]

        assert wild == wild.edges["*"]
        assert wild.final


class TestTokenSetFromList:
    def test_from_list_with_unsorted_list(self):
        with pytest.raises(BaseLunrException):
            TokenSet.from_list(["z", "a"])

    def test_from_list_with_sorted_list(self):
        token_set = TokenSet.from_list(["a", "z"])
        assert ["a", "z"] == sorted(token_set.to_list())

    def test_from_list_is_minimal(self):
        token_set = TokenSet.from_list(["ac", "dc"])
        ac_node = token_set.edges["a"].edges["c"]
        dc_node = token_set.edges["d"].edges["c"]

        assert ac_node == dc_node


class TestTokenSetToList:
    def test_to_list_includes_all_words(self):
        words = ["bat", "cat"]
        token_set = TokenSet.from_list(words)

        assert set(words) == set(token_set.to_list())

    def test_to_list_includes_single_words(self):
        word = "bat"
        token_set = TokenSet.from_string(word)

        assert {word} == set(token_set.to_list())


class TestTokenSetIntersect:
    def test_no_intersection(self):
        x = TokenSet.from_string("cat")
        y = TokenSet.from_string("bar")
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_simple_intersection(self):
        x = TokenSet.from_string("cat")
        y = TokenSet.from_string("cat")
        z = x.intersect(y)

        assert {"cat"} == set(z.to_list())

    def test_trailing_wildcard_intersection(self):
        x = TokenSet.from_string("cat")
        y = TokenSet.from_string("c*")
        z = x.intersect(y)

        assert {"cat"} == set(z.to_list())

    def test_trailing_wildcard_no_intersection(self):
        x = TokenSet.from_string("cat")
        y = TokenSet.from_string("b*")
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_leading_wildcard_intersection(self):
        x = TokenSet.from_string("cat")
        y = TokenSet.from_string("*t")
        z = x.intersect(y)

        assert {"cat"} == set(z.to_list())

    def test_leading_wildcard_no_intersection(self):
        x = TokenSet.from_string("cat")
        y = TokenSet.from_string("*r")
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_contained_wildcard_intersection(self):
        x = TokenSet.from_string("foo")
        y = TokenSet.from_string("f*o")
        z = x.intersect(y)

        assert {"foo"} == set(z.to_list())

    def test_contained_wildcard_no_intersection(self):
        x = TokenSet.from_string("foo")
        y = TokenSet.from_string("b*r")
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_wildcard_zero_or_more_characters(self):
        x = TokenSet.from_string("foo")
        y = TokenSet.from_string("foo*")
        z = x.intersect(y)

        assert {"foo"} == set(z.to_list())

    def test_with_fuzzy_string_substitution(self):
        x1 = TokenSet.from_string("bar")
        x2 = TokenSet.from_string("cur")
        x3 = TokenSet.from_string("cat")
        x4 = TokenSet.from_string("car")
        x5 = TokenSet.from_string("foo")
        y = TokenSet.from_fuzzy_string("car", 1)

        assert x1.intersect(y).to_list() == ["bar"]
        assert x2.intersect(y).to_list() == ["cur"]
        assert x3.intersect(y).to_list() == ["cat"]
        assert x4.intersect(y).to_list() == ["car"]
        assert x5.intersect(y).to_list() == []

    def test_with_fuzzy_string_deletion(self):
        x1 = TokenSet.from_string("ar")
        x2 = TokenSet.from_string("br")
        x3 = TokenSet.from_string("ba")
        x4 = TokenSet.from_string("bar")
        x5 = TokenSet.from_string("foo")
        y = TokenSet.from_fuzzy_string("bar", 1)

        assert x1.intersect(y).to_list() == ["ar"]
        assert x2.intersect(y).to_list() == ["br"]
        assert x3.intersect(y).to_list() == ["ba"]
        assert x4.intersect(y).to_list() == ["bar"]
        assert x5.intersect(y).to_list() == []

    def test_with_fuzzy_string_insertion(self):
        x1 = TokenSet.from_string("bbar")
        x2 = TokenSet.from_string("baar")
        x3 = TokenSet.from_string("barr")
        x4 = TokenSet.from_string("bar")
        x5 = TokenSet.from_string("ba")
        x6 = TokenSet.from_string("foo")
        x7 = TokenSet.from_string("bara")
        y = TokenSet.from_fuzzy_string("bar", 1)

        assert x1.intersect(y).to_list() == ["bbar"]
        assert x2.intersect(y).to_list() == ["baar"]
        assert x3.intersect(y).to_list() == ["barr"]
        assert x4.intersect(y).to_list() == ["bar"]
        assert x5.intersect(y).to_list() == ["ba"]
        assert x6.intersect(y).to_list() == []
        assert x7.intersect(y).to_list() == ["bara"]

    def test_with_fuzzy_string_transpose(self):
        x1 = TokenSet.from_string("abr")
        x2 = TokenSet.from_string("bra")
        x3 = TokenSet.from_string("foo")
        y = TokenSet.from_fuzzy_string("bar", 1)

        assert x1.intersect(y).to_list() == ["abr"]
        assert x2.intersect(y).to_list() == ["bra"]
        assert x3.intersect(y).to_list() == []

    def test_fuzzy_string_insertion(self):
        x = TokenSet.from_string("abcxx")
        y = TokenSet.from_fuzzy_string("abc", 2)

        assert x.intersect(y).to_list() == ["abcxx"]

    def test_fuzzy_string_substitution(self):
        x = TokenSet.from_string("axx")
        y = TokenSet.from_fuzzy_string("abc", 2)

        assert x.intersect(y).to_list() == ["axx"]

    def test_fuzzy_string_deletion(self):
        x = TokenSet.from_string("a")
        y = TokenSet.from_fuzzy_string("abc", 2)

        assert x.intersect(y).to_list() == ["a"]

    def test_fuzzy_string_transpose(self):
        x = TokenSet.from_string("bca")
        y = TokenSet.from_fuzzy_string("abc", 2)

        assert x.intersect(y).to_list() == ["bca"]

    def test_leading_wildcard_backtracking_intersection(self):
        x = TokenSet.from_string("aaacbab")
        y = TokenSet.from_string("*ab")

        assert x.intersect(y).to_list() == ["aaacbab"]

    def test_leading_wildcard_backtracking_no_intersection(self):
        x = TokenSet.from_string("aaacbab")
        y = TokenSet.from_string("*abc")

        assert x.intersect(y).to_list() == []

    def test_contained_wildcard_backtracking_intersection(self):
        x = TokenSet.from_string("ababc")
        y = TokenSet.from_string("a*bc")

        assert x.intersect(y).to_list() == ["ababc"]

    def test_contained_wildcard_backtracking_no_intersection(self):
        x = TokenSet.from_string("ababc")
        y = TokenSet.from_string("a*ac")

        assert x.intersect(y).to_list() == []

    @pytest.mark.timeout(2)
    def test_catastrophic_backtracking_with_leading_characters(self):
        x = TokenSet.from_string("f" * 100)
        y = TokenSet.from_string("*f")

        assert len(x.intersect(y).to_list()) == 1

    def test_leading_trailing_wildcard_backtracking_intersection(self):
        x = TokenSet.from_string("acbaabab")
        y = TokenSet.from_string("*ab*")

        assert x.intersect(y).to_list() == ["acbaabab"]

    def test_leading_atrailing_wildcard_backtracking_intersection(self):
        x = TokenSet.from_string("acbaabab")
        y = TokenSet.from_string("a*ba*b")

        assert x.intersect(y).to_list() == ["acbaabab"]



import os
import re
from setuptools import setup, find_packages


PATH = os.path.abspath(os.path.dirname(__file__))


def read_file(filepath):
    with open(filepath, "r") as fd:
        return fd.read()


def find_version():
    version_path = os.path.join(PATH, "lunr", "__init__.py")
    contents = read_file(version_path)
    version_string = contents[contents.index("__VERSION__") :]
    try:
        return re.match(r'.*__VERSION__ = [\'"]([\d\w\.]+)[\'"]', version_string).group(
            1
        )
    except AttributeError:
        raise RuntimeError("Unable to find version string.")


setup(
    name="lunr",
    version=find_version(),
    url="https://github.com/yeraydiazdiaz/lunr.py",
    project_urls={"Documentation": "https://lunr.readthedocs.io"},
    license="MIT",
    description="A Python implementation of Lunr.js",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    author="Yeray Diaz Diaz",
    author_email="yeraydiazdiaz@gmail.com",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=[],
    extras_require={
        "languages": ["nltk"]
    },
    keywords="lunr full text search",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Text Processing",
    ],
)



import json
from unittest.mock import MagicMock, patch

import pytest

from lunr import __TARGET_JS_VERSION__
from lunr.index import Index
from lunr.exceptions import BaseLunrException

from tests.utils import assert_vectors_equal


class TestIndex:
    def test_create_query_default_fields(self, index):
        query = index.create_query()
        assert query.all_fields == index.fields

    def test_create_query_subset_of_fields(self, index):
        query = index.create_query([index.fields[0]])
        assert query.all_fields == [index.fields[0]]

    def test_create_query_non_contained_fields(self, index):
        with pytest.raises(BaseLunrException):
            index.create_query(["foo"])

    def test_query_no_arguments_warns_and_returns_no_results(self, monkeypatch, index):
        from lunr.index import logger

        mock_logger = MagicMock()
        monkeypatch.setattr(logger, "warning", mock_logger)
        results = index.query()
        assert results == []
        mock_logger.assert_called_once()

    def test_query_callback_argument_is_query_with_fields(self, index):
        def callback(query):
            assert query.all_fields == index.fields

        index.query(callback=callback)

    def test_query_callback_can_configure_query(self, index):
        def callback(query):
            query.clause("study")

        results = index.query(callback=callback)
        assert len(results) == 2
        assert results[0]["ref"] == "b"
        assert results[1]["ref"] == "a"


class TestIndexSerialization:
    def test_serialization(self, index):
        serialized_index = index.serialize()
        assert serialized_index["version"] == __TARGET_JS_VERSION__
        assert serialized_index["fields"] == index.fields
        for ref, vector in serialized_index["fieldVectors"]:
            assert ref in index.field_vectors
            assert_vectors_equal(vector, index.field_vectors[ref])

    def test_json_deserialization(self, index):
        serialized_index = index.serialize()
        json_serialized_index = json.dumps(serialized_index)

        idx = Index.load(json_serialized_index)

        assert idx == index

    def test_load_warns_on_js_version_mismatch(self, index):
        serialized_index = index.serialize()
        serialized_index["version"] = "1.0.0"

        with patch("lunr.index.logger") as mock_log:
            Index.load(serialized_index)
            mock_log.warning.assert_called_once()



from lunr import languages as lang
from lunr.builder import Builder
from lunr.stemmer import stemmer
from lunr.trimmer import trimmer
from lunr.stop_word_filter import stop_word_filter


def lunr(ref, fields, documents, languages=None, builder=None):
    """A convenience function to configure and construct a lunr.Index.

    Args:
        ref (str): The key in the documents to be used a the reference.
        fields (list): A list of strings defining fields in the documents to
            index. Optionally a list of dictionaries with three keys:
            `field_name` defining the document's field, `boost` an integer
            defining a boost to be applied to the field, and `extractor`
            a callable taking the document as a single argument and returning
            a string located in the document in a particular way.
        documents (list): The list of dictonaries representing the documents
            to index. Optionally a 2-tuple of dicts, the first one being
            the document and the second the associated attributes to it.
        languages (str or list, optional): The languages to use if using
            NLTK language support, ignored if NLTK is not available.

    Returns:
        Index: The populated Index ready to search against.
    """
    builder = builder or get_default_builder(languages)
    builder.ref(ref)
    for field in fields:
        if isinstance(field, dict):
            builder.field(**field)
        else:
            builder.field(field)

    for document in documents:
        if isinstance(document, (tuple, list)):
            builder.add(document[0], attributes=document[1])
        else:
            builder.add(document)

    return builder.build()


def get_default_builder(languages=None):
    """Creates a new pre-configured instance of Builder.

    Useful as a starting point to tweak the defaults.
    """
    if languages is not None and lang.LANGUAGE_SUPPORT:
        if isinstance(languages, str):
            languages = [languages]

        unsupported_languages = set(languages) - set(lang.SUPPORTED_LANGUAGES)
        if unsupported_languages:
            raise RuntimeError(
                "The specified languages {} are not supported, "
                "please choose one of {}".format(
                    ", ".join(unsupported_languages),
                    ", ".join(lang.SUPPORTED_LANGUAGES.keys()),
                )
            )
        builder = lang.get_nltk_builder(languages)
    else:
        builder = Builder()
        builder.pipeline.add(trimmer, stop_word_filter, stemmer)
        builder.search_pipeline.add(stemmer)

    return builder



import pytest

from lunr.query import Query, QueryPresence
from lunr.token import Token
from lunr.tokenizer import Tokenizer


class BaseQuerySuite:
    ALL_FIELDS = ["title", "body"]

    def setup_method(self, method):
        self.query = Query(self.ALL_FIELDS)


class TestQueryTerm(BaseQuerySuite):
    def test_single_string_term_adds_single_clause(self):
        self.query.clause(term="foo")

        assert len(self.query.clauses) == 1
        assert self.query.clauses[0].term == "foo"
        assert repr(self.query) == '<Query fields="title,body" clauses="foo">'
        assert repr(self.query.clauses[0]) == '<Clause term="foo">'

    def test_single_token_term_adds_single_clause(self):
        self.query.term(Token("foo"))

        assert len(self.query.clauses) == 1
        assert self.query.clauses[0].term == "foo"

    def test_multiple_string_terms_adds_multiple_clauses(self):
        self.query.term(["foo", "bar"])

        assert len(self.query.clauses) == 2
        assert self.query.clauses[0].term == "foo"
        assert self.query.clauses[1].term == "bar"
        assert repr(self.query) == ('<Query fields="title,body" clauses="foo,bar">')

    def test_multiple_token_terms_adds_multiple_clauses(self):
        self.query.term(Tokenizer("foo bar"))

        assert len(self.query.clauses) == 2
        assert self.query.clauses[0].term == "foo"
        assert self.query.clauses[1].term == "bar"

    def test_multiple_string_terms_with_options(self):
        self.query.term(["foo", "bar"], use_pipeline=False)

        assert len(self.query.clauses) == 2
        assert self.query.clauses[0].term == "foo"
        assert self.query.clauses[1].term == "bar"


class TestQueryClause(BaseQuerySuite):
    def test_clause_defaults(self):
        self.query.clause(term="foo")
        self.clause = self.query.clauses[0]

        assert self.clause.fields == self.ALL_FIELDS
        assert self.clause.boost == 1
        assert self.clause.use_pipeline is True

    def test_clause_specified(self):
        self.query.clause(term="foo", boost=10, fields=["title"], use_pipeline=False)
        self.clause = self.query.clauses[0]

        assert self.clause.fields == ["title"]
        assert self.clause.boost == 10
        assert self.clause.use_pipeline is False

    @pytest.mark.parametrize(
        "wildcard, expected_term",
        [
            (Query.WILDCARD_NONE, "foo"),
            (Query.WILDCARD_LEADING, "*foo"),
            (Query.WILDCARD_TRAILING, "foo*"),
            (Query.WILDCARD_LEADING | Query.WILDCARD_TRAILING, "*foo*"),
        ],
    )
    def test_clause_wildcard(self, wildcard, expected_term):
        self.query.clause(term="foo", wildcard=wildcard)
        self.clause = self.query.clauses[0]

        assert self.clause.term == expected_term

    def test_clause_wildcard_existing(self):
        self.query.clause(
            term="*foo*", wildcard=Query.WILDCARD_LEADING | Query.WILDCARD_TRAILING
        )
        self.clause = self.query.clauses[0]

        assert self.clause.term == "*foo*"


class TestQueryIsNegated(BaseQuerySuite):
    def test_all_prohibited(self):
        self.query.term("foo", presence=QueryPresence.PROHIBITED)
        self.query.term("bar", presence=QueryPresence.PROHIBITED)

        assert self.query.is_negated() is True

    def test_some_prohibited(self):
        self.query.term("foo", presence=QueryPresence.PROHIBITED)
        self.query.term("bar", presence=QueryPresence.REQUIRED)

        assert self.query.is_negated() is False

    def test_nome_prohibited(self):
        self.query.term("foo", presence=QueryPresence.OPTIONAL)
        self.query.term("bar", presence=QueryPresence.REQUIRED)

        assert self.query.is_negated() is False



from collections import defaultdict
import json
import logging

from lunr.exceptions import BaseLunrException
from lunr.field_ref import FieldRef
from lunr.match_data import MatchData
from lunr.token_set import TokenSet
from lunr.token_set_builder import TokenSetBuilder
from lunr.pipeline import Pipeline
from lunr.query import Query, QueryPresence
from lunr.query_parser import QueryParser
from lunr.utils import CompleteSet
from lunr.vector import Vector

logger = logging.getLogger(__name__)


class Index:
    """An index contains the built index of all documents and provides a query
    interface to the index.

    Usually instances of lunr.Index will not be created using this
    constructor, instead lunr.Builder should be used to construct new
    indexes, or lunr.Index.load should be used to load previously built and
    serialized indexes.
    """

    def __init__(self, inverted_index, field_vectors, token_set, fields, pipeline):
        self.inverted_index = inverted_index
        self.field_vectors = field_vectors
        self.token_set = token_set
        self.fields = fields
        self.pipeline = pipeline

    def __eq__(self, other):
        # TODO: extend equality to other attributes
        return (
            self.inverted_index == other.inverted_index and self.fields == other.fields
        )

    def search(self, query_string):
        """Performs a search against the index using lunr query syntax.

        Results will be returned sorted by their score, the most relevant
        results will be returned first.

        For more programmatic querying use `lunr.Index.query`.

        Args:
            query_string (str): A string to parse into a Query.

        Returns:
            dict: Results of executing the query.
        """
        query = self.create_query()
        # TODO: should QueryParser be a method of query? should it return one?
        parser = QueryParser(query_string, query)
        parser.parse()
        return self.query(query)

    def create_query(self, fields=None):
        """Convenience method to create a Query with the Index's fields.

        Args:
            fields (iterable, optional): The fields to include in the Query,
                defaults to the Index's `all_fields`.

        Returns:
            Query: With the specified fields or all the fields in the Index.
        """
        if fields is None:
            return Query(self.fields)

        non_contained_fields = set(fields) - set(self.fields)
        if non_contained_fields:
            raise BaseLunrException(
                "Fields {} are not part of the index", non_contained_fields
            )

        return Query(fields)

    def query(self, query=None, callback=None):
        """Performs a query against the index using the passed lunr.Query
        object.

        If performing programmatic queries against the index, this method is
        preferred over `lunr.Index.search` so as to avoid the additional query
        parsing overhead.

        Args:
            query (lunr.Query): A configured Query to perform the search
                against, use `create_query` to get a preconfigured object
                or use `callback` for convenience.
            callback (callable): An optional function taking a single Query
                object result of `create_query` for further configuration.
        """
        if query is None:
            query = self.create_query()

        if callback is not None:
            callback(query)

        if len(query.clauses) == 0:
            logger.warning(
                "Attempting a query with no clauses. Please add clauses by "
                "either using the `callback` argument or using `create_query` "
                "to create a preconfigured Query, manually adding clauses and "
                "passing it as the `query` argument."
            )
            return []

        # for each query clause
        # * process terms
        # * expand terms from token set
        # * find matching documents and metadata
        # * get document vectors
        # * score documents

        matching_fields = {}
        query_vectors = {field: Vector() for field in self.fields}
        term_field_cache = {}
        required_matches = {}
        prohibited_matches = defaultdict(set)

        for clause in query.clauses:
            # Unless the pipeline has been disabled for this term, which is
            # the case for terms with wildcards, we need to pass the clause
            # term through the search pipeline. A pipeline returns an array
            # of processed terms. Pipeline functions may expand the passed
            # term, which means we may end up performing multiple index lookups
            # for a single query term.
            if clause.use_pipeline:
                terms = self.pipeline.run_string(clause.term, {"fields": clause.fields})
            else:
                terms = [clause.term]

            clause_matches = set()

            for term in terms:
                # Each term returned from the pipeline needs to use the same
                # query clause object, e.g. the same boost and or edit distance
                # The simplest way to do this is to re-use the clause object
                # but mutate its term property.
                clause.term = term

                # From the term in the clause we create a token set which will
                # then be used to intersect the indexes token set to get a list
                # of terms to lookup in the inverted index
                term_token_set = TokenSet.from_clause(clause)
                expanded_terms = self.token_set.intersect(term_token_set).to_list()

                # If a term marked as required does not exist in the TokenSet
                # it is impossible for the search to return any matches.
                # We set all the field-scoped required matches set to empty
                # and stop examining further clauses
                if (
                    len(expanded_terms) == 0
                    and clause.presence == QueryPresence.REQUIRED
                ):
                    for field in clause.fields:
                        required_matches[field] = CompleteSet()
                    break

                for expanded_term in expanded_terms:
                    posting = self.inverted_index[expanded_term]
                    term_index = posting["_index"]

                    for field in clause.fields:
                        # For each field that this query term is scoped by
                        # (by default all fields are in scope) we need to get
                        # all the document refs that have this term in that
                        # field.
                        #
                        # The posting is the entry in the invertedIndex for the
                        # matching term from above.
                        field_posting = posting[field]
                        matching_document_refs = field_posting.keys()
                        term_field = expanded_term + "/" + field
                        matching_documents_set = set(matching_document_refs)

                        # If the presence of this term is required, ensure that
                        # the matching documents are added to the set of
                        # required matches for this clause.
                        if clause.presence == QueryPresence.REQUIRED:
                            clause_matches = clause_matches.union(
                                matching_documents_set
                            )

                            if field not in required_matches:
                                required_matches[field] = CompleteSet()

                        # If the presence of this term is prohibited,
                        # ensure that the matching documents are added to the
                        # set of prohibited matches for this field, creating
                        # that set if it does not exist yet.
                        elif clause.presence == QueryPresence.PROHIBITED:
                            prohibited_matches[field] = prohibited_matches[field].union(
                                matching_documents_set
                            )

                            # prohibited matches should not be part of the
                            # query vector used for similarity scoring and no
                            # metadata should be extracted so we continue
                            # to the next field
                            continue

                        # The query field vector is populated using the
                        # term_index found for the term an a unit value with
                        # the appropriate boost
                        # Using upsert because there could already be an entry
                        # in the vector for the term we are working with.
                        # In that case we just add the scores together.
                        query_vectors[field].upsert(
                            term_index, clause.boost, lambda a, b: a + b
                        )

                        # If we've already seen this term, field combo then
                        # we've already collected the matching documents and
                        # metadata, no need to go through all that again
                        if term_field in term_field_cache:
                            continue

                        for matching_document_ref in matching_document_refs:
                            # All metadata for this term/field/document triple
                            # are then extracted and collected into an instance
                            # of lunr.MatchData ready to be returned in the
                            # query results
                            matching_field_ref = FieldRef(matching_document_ref, field)
                            metadata = field_posting[str(matching_document_ref)]

                            if str(matching_field_ref) not in matching_fields:
                                matching_fields[str(matching_field_ref)] = MatchData(
                                    expanded_term, field, metadata
                                )
                            else:
                                matching_fields[str(matching_field_ref)].add(
                                    expanded_term, field, metadata
                                )

                        term_field_cache[term_field] = True

            # if the presence was required we need to update the required
            # matches field sets, we do this after all fields for the term
            # have collected their matches because the clause terms presence
            # is required in _any_ of the fields, not _all_ of the fields
            if clause.presence == QueryPresence.REQUIRED:
                for field in clause.fields:
                    required_matches[field] = required_matches[field].intersection(
                        clause_matches
                    )

        # We need to combine the field scoped required and prohibited
        # matching documents inot a global set of required and prohibited
        # matches
        all_required_matches = CompleteSet()
        all_prohibited_matches = set()
        for field in self.fields:
            if field in required_matches:
                all_required_matches = all_required_matches.intersection(
                    required_matches[field]
                )
            if field in prohibited_matches:
                all_prohibited_matches = all_prohibited_matches.union(
                    prohibited_matches[field]
                )

        matching_field_refs = matching_fields.keys()
        results = []
        matches = {}

        # If the query is negated (only contains prohibited terms)
        # we need to get _all_ field_refs currently existing in the index.
        # This to avoid any costs of getting all field regs unnecessarily
        # Additionally, blank match data must be created to correctly populate
        # the results
        if query.is_negated():
            matching_field_refs = list(self.field_vectors.keys())

            for matching_field_ref in matching_field_refs:
                field_ref = FieldRef.from_string(matching_field_ref)
                matching_fields[matching_field_ref] = MatchData()

        for matching_field_ref in matching_field_refs:
            # Currently we have document fields that match the query, but we
            # need to return documents. The matchData and scores are combined
            # from multiple fields belonging to the same document.
            #
            # Scores are calculated by field, using the query vectors created
            # above, and combined into a final document score using addition.
            field_ref = FieldRef.from_string(matching_field_ref)
            doc_ref = field_ref.doc_ref

            if doc_ref not in all_required_matches or doc_ref in all_prohibited_matches:
                continue

            field_vector = self.field_vectors[matching_field_ref]
            score = query_vectors[field_ref.field_name].similarity(field_vector)

            try:
                doc_match = matches[doc_ref]
                doc_match["score"] += score
                doc_match["match_data"].combine(matching_fields[matching_field_ref])
            except KeyError:
                match = {
                    "ref": doc_ref,
                    "score": score,
                    "match_data": matching_fields[matching_field_ref],
                }
                matches[doc_ref] = match
                results.append(match)

        return sorted(results, key=lambda a: a["score"], reverse=True)

    def serialize(self):
        """Returns a serialized index as a dict following lunr-schema."""
        from lunr import __TARGET_JS_VERSION__

        inverted_index = [
            [term, self.inverted_index[term]] for term in sorted(self.inverted_index)
        ]
        field_vectors = [
            [ref, vector.serialize()] for ref, vector in self.field_vectors.items()
        ]

        # CamelCased keys for compatibility with JS version
        return {
            "version": __TARGET_JS_VERSION__,
            "fields": self.fields,
            "fieldVectors": field_vectors,
            "invertedIndex": inverted_index,
            "pipeline": self.pipeline.serialize(),
        }

    @classmethod
    def load(cls, serialized_index):
        """Load a serialized index"""
        from lunr import __TARGET_JS_VERSION__

        if isinstance(serialized_index, str):
            serialized_index = json.loads(serialized_index)

        if serialized_index["version"] != __TARGET_JS_VERSION__:
            logger.warning(
                "Version mismatch when loading serialized index. "
                "Current version of lunr {} does not match that of serialized "
                "index {}".format(__TARGET_JS_VERSION__, serialized_index["version"])
            )

        field_vectors = {
            ref: Vector(elements) for ref, elements in serialized_index["fieldVectors"]
        }

        tokenset_builder = TokenSetBuilder()
        inverted_index = {}
        for term, posting in serialized_index["invertedIndex"]:
            tokenset_builder.insert(term)
            inverted_index[term] = posting

        tokenset_builder.finish()

        return Index(
            fields=serialized_index["fields"],
            field_vectors=field_vectors,
            inverted_index=inverted_index,
            token_set=tokenset_builder.root,
            pipeline=Pipeline.load(serialized_index["pipeline"]),
        )



import json
import os

from lunr.token import Token
from lunr.stemmer import stemmer
from lunr.pipeline import Pipeline


class TestStemmer:
    def test_reduces_words_to_their_stem(self):
        path = os.path.join(
            os.path.dirname(__file__), "fixtures", "stemming_vocab.json"
        )
        with open(path) as f:
            data = json.loads(f.read())

        for word, expected in data.items():
            token = Token(word)
            result = str(stemmer(token))

            assert result == expected

    def test_is_a_registered_pipeline_function(self):
        assert stemmer.label == "stemmer"
        assert Pipeline.registered_functions["stemmer"] == stemmer



import math


def idf(posting, document_count):
    """A function to calculate the inverse document frequency for a posting.
    This is shared between the builder and the index.
    """
    documents_with_term = 0
    for field_name in posting:
        if field_name == "_index":
            continue
        documents_with_term += len(posting[field_name].keys())

    x = (document_count - documents_with_term + 0.5) / (documents_with_term + 0.5)
    return math.log(1 + abs(x))



from lunr.token_set import TokenSet
from lunr.exceptions import BaseLunrException


class TokenSetBuilder:
    def __init__(self):
        self.previous_word = ""
        self.root = TokenSet()
        self.unchecked_nodes = []
        self.minimized_nodes = {}

    def insert(self, word):
        if word < self.previous_word:
            raise BaseLunrException("Out of order word insertion")

        common_prefix = 0
        for i in range(min(len(word), len(self.previous_word))):
            if word[i] != self.previous_word[i]:
                break

            common_prefix += 1

        self.minimize(common_prefix)

        node = (
            self.root if not self.unchecked_nodes else self.unchecked_nodes[-1]["child"]
        )

        for i in range(common_prefix, len(word)):
            next_node = TokenSet()
            char = word[i]

            node.edges[char] = next_node

            self.unchecked_nodes.append(
                {"parent": node, "char": char, "child": next_node}
            )

            node = next_node

        node.final = True
        self.previous_word = word

    def finish(self):
        self.minimize(0)

    def minimize(self, down_to):
        for i in range(len(self.unchecked_nodes) - 1, down_to - 1, -1):
            node = self.unchecked_nodes[i]
            child_key = str(node["child"])

            if child_key in self.minimized_nodes:
                node["parent"].edges[node["char"]] = self.minimized_nodes[child_key]
            else:
                node["child"]._str = child_key
                self.minimized_nodes[child_key] = node["child"]

            self.unchecked_nodes.pop()



class TokenSet:
    """
    A token set is used to store the unique list of all tokens
    within an index. Token sets are also used to represent an
    incoming query to the index, this query token set and index
    token set are then intersected to find which tokens to look
    up in the inverted index.

    A token set can hold multiple tokens, as in the case of the
    index token set, or it can hold a single token as in the
    case of a simple query token set.

    Additionally token sets are used to perform wildcard matching.
    Leading, contained and trailing wildcards are supported, and
    from this edit distance matching can also be provided.

    Token sets are implemented as a minimal finite state automata,
    where both common prefixes and suffixes are shared between tokens.
    This helps to reduce the space used for storing the token set.

    TODO: consider https://github.com/glyph/automat
    """

    _next_id = 1

    def __init__(self):
        self.final = False
        self.edges = {}
        self.id = self._next_id
        self.__class__._next_id += 1

    def __str__(self):
        try:
            return self._string
        except AttributeError:
            pass

        string = "1" if self.final else "0"
        for label in sorted(list(self.edges.keys())):
            node = self.edges[label]
            try:
                node_id = str(node.id)
            except AttributeError:
                # TODO: JS seems to rely on undefined for the id attribute?
                node_id = ""

            string = string + label + node_id

        return string

    def __repr__(self):
        return '<TokenSet "{}">'.format(str(self))

    @classmethod
    def from_string(self, string):
        """Creates a TokenSet from a string.

        The string may contain one or more wildcard characters (*) that will
        allow wildcard matching when intersecting with another TokenSet
        """
        node = TokenSet()
        root = node

        # Iterates throough all characters in the passed string appending
        # a node for each character.
        # When a wildcard character is found then a self referencing edge
        # is introduced to continually match any number of characters
        for i, char in enumerate(string):
            final = i == len(string) - 1
            if char == "*":
                node.edges[char] = node
                node.final = final
            else:
                next_ = TokenSet()
                next_.final = final
                node.edges[char] = next_
                node = next_

        return root

    @classmethod
    def from_fuzzy_string(cls, string, edit_distance):
        """Creates a token set representing a single string with a specified
        edit distance.

        Insertions, deletions, substitutions and transpositions are each
        treated as an edit distance of 1.

        Increasing the allowed edit distance will have a dramatic impact
        on the performance of both creating and intersecting these TokenSets.
        It is advised to keep the edit distance less than 3.
        """
        root = TokenSet()

        stack = [{"node": root, "edits_remaining": edit_distance, "string": string}]

        while stack:
            frame = stack.pop()
            # no edit
            if len(frame["string"]) > 0:
                char = frame["string"][0]
                no_edit_node = None
                if char in frame["node"].edges:
                    no_edit_node = frame["node"].edges[char]
                else:
                    no_edit_node = TokenSet()
                    frame["node"].edges[char] = no_edit_node

                if len(frame["string"]) == 1:
                    no_edit_node.final = True

                stack.append(
                    {
                        "node": no_edit_node,
                        "edits_remaining": frame["edits_remaining"],
                        "string": frame["string"][1:],
                    }
                )

            if frame["edits_remaining"] == 0:
                continue

            # insertion, can only do insertion if there are edits remaining
            if "*" in frame["node"].edges:
                insertion_node = frame["node"].edges["*"]
            else:
                insertion_node = TokenSet()
                frame["node"].edges["*"] = insertion_node

            if len(frame["string"]) == 0:
                insertion_node.final = True

            stack.append(
                {
                    "node": insertion_node,
                    "edits_remaining": frame["edits_remaining"] - 1,
                    "string": frame["string"],
                }
            )

            # deletion, can only do a deletion if we have enough edits
            # remaining and if there are characters left to delete in the string
            if len(frame["string"]) > 1:
                stack.append(
                    {
                        "node": frame["node"],
                        "edits_remaining": frame["edits_remaining"] - 1,
                        "string": frame["string"][1:],
                    }
                )

            # deletion, just removing the last character of the string
            if len(frame["string"]) == 1:
                frame["node"].final = True

            # substitution, can only do a substitution if we have enough edits
            # remaining and there are characters left to substitute
            if len(frame["string"]) >= 1:
                if "*" in frame["node"].edges:
                    substitution_node = frame["node"].edges["*"]
                else:
                    substitution_node = TokenSet()
                    frame["node"].edges["*"] = substitution_node

                if len(frame["string"]) == 1:
                    substitution_node.final = True

                stack.append(
                    {
                        "node": substitution_node,
                        "edits_remaining": frame["edits_remaining"] - 1,
                        "string": frame["string"][1:],
                    }
                )

            # transposition, can only do a transposition if there are edits
            # remaining and there are enough characters to transpose
            if frame["edits_remaining"] and len(frame["string"]) > 1:
                char_a = frame["string"][0]
                char_b = frame["string"][1]
                transpose_node = None

                if char_b in frame["node"].edges:
                    transpose_node = frame["node"].edges[char_b]
                else:
                    transpose_node = TokenSet()
                    frame["node"].edges[char_b] = transpose_node

                if len(frame["string"]) == 1:
                    transpose_node.final = True

                stack.append(
                    {
                        "node": transpose_node,
                        "edits_remaining": frame["edits_remaining"] - 1,
                        "string": char_a + frame["string"][2:],
                    }
                )

        return root

    @classmethod
    def from_list(cls, list_of_words):
        from lunr.token_set_builder import TokenSetBuilder

        builder = TokenSetBuilder()
        for word in list_of_words:
            builder.insert(word)

        builder.finish()
        return builder.root

    @classmethod
    def from_clause(cls, clause):
        if clause.edit_distance:
            return cls.from_fuzzy_string(clause.term, clause.edit_distance)
        else:
            return cls.from_string(clause.term)

    def to_list(self):
        words = []
        stack = [{"prefix": "", "node": self}]

        while stack:
            frame = stack.pop()
            if frame["node"].final:
                words.append(frame["prefix"])

            for edge in frame["node"].edges.keys():
                stack.append(
                    {
                        "prefix": frame["prefix"] + str(edge),
                        "node": frame["node"].edges[edge],
                    }
                )

        return words

    def intersect(self, other):
        """Returns a new TokenSet that is the intersection of this TokenSet
        and the passed TokenSet.

        This intersection will take into account any wildcards contained within
        the TokenSet.
        """
        output = TokenSet()
        stack = [{"node": self, "q_node": other, "output": output}]

        while stack:
            frame = stack.pop()
            for q_edge in frame["q_node"].edges.keys():
                for n_edge in frame["node"].edges.keys():
                    if n_edge == q_edge or q_edge == "*":
                        node = frame["node"].edges[n_edge]
                        q_node = frame["q_node"].edges[q_edge]
                        final = node.final and q_node.final
                        next_ = None

                        if n_edge in frame["output"].edges:
                            next_ = frame["output"].edges[n_edge]
                            next_.final = next_.final or final
                        else:
                            next_ = TokenSet()
                            next_.final = final
                            frame["output"].edges[n_edge] = next_

                        stack.append({"node": node, "q_node": q_node, "output": next_})

        return output



import os.path

import pytest

from tests.utils import read_json_fixture

from lunr import lunr
from lunr.pipeline import Pipeline


def get_mkdocs_index():
    data = read_json_fixture("mkdocs_index.json")
    return lunr(ref="id", fields=("title", "text"), documents=data["docs"])


class TestSearchBenchmarks:
    @pytest.fixture(scope="session")
    def index(self):
        return get_mkdocs_index()

    def test_search(self, index, benchmark):
        benchmark(index.search, "styling")


class TestPipelineBenchmarks:
    FEW_COUNT = 50
    MANY_COUNT = 1000

    @pytest.fixture(scope="session")
    def many_tokens(self):
        path = os.path.join(os.path.dirname(__file__), "fixtures/words.txt")
        with open(path) as words:
            self.many_tokens = [
                words.readline().strip() for _ in range(self.MANY_COUNT)
            ]
        self.few_tokens = self.many_tokens[: self.FEW_COUNT]
        yield self.many_tokens

    @pytest.fixture(scope="session")
    def few_tokens(self, many_tokens):
        yield self.few_tokens

    @staticmethod
    def token_to_token(token, i, tokens):
        return token

    @staticmethod
    def token_to_token_array(token, i, tokens):
        return [token, token]

    def test_few_token_to_token(self, few_tokens, benchmark):
        token_to_token_pipeline = Pipeline()
        token_to_token_pipeline.add(self.token_to_token)
        benchmark(token_to_token_pipeline.run, few_tokens)

    def test_many_token_to_token(self, many_tokens, benchmark):
        token_to_token_pipeline = Pipeline()
        token_to_token_pipeline.add(self.token_to_token)
        benchmark(token_to_token_pipeline.run, many_tokens)

    def test_few_token_to_token_array(self, few_tokens, benchmark):
        token_to_token_array_pipeline = Pipeline()
        token_to_token_array_pipeline.add(self.token_to_token_array)
        benchmark(token_to_token_array_pipeline.run, few_tokens)

    def test_many_token_to_token_array(self, many_tokens, benchmark):
        token_to_token_array_pipeline = Pipeline()
        token_to_token_array_pipeline.add(self.token_to_token_array)
        benchmark(token_to_token_array_pipeline.run, many_tokens)


if __name__ == "__main__":
    get_mkdocs_index()



from lunr.tokenizer import default_separator


class QueryLexer:
    # TODO: use iteration protocol?
    EOS = "EOS"
    FIELD = "FIELD"
    TERM = "TERM"
    EDIT_DISTANCE = "EDIT_DISTANCE"
    BOOST = "BOOST"
    PRESENCE = "PRESENCE"

    def __init__(self, string):
        self.lexemes = []
        self.string = string
        self.length = len(string)
        self.pos = 0
        self.start = 0
        self.escape_char_positions = []

    @property
    def width(self):
        return self.pos - self.start

    def ignore(self):
        if self.start == self.pos:
            self.pos += 1

        self.start = self.pos

    def backup(self):
        self.pos -= 1

    def accept_digit_run(self):
        char = self.next()
        while char != self.EOS and (47 < ord(char) < 58):
            char = self.next()

        if char != self.EOS:
            self.backup()

    def run(self):
        state = self.lex_text()
        while state:
            state = state()

    def slice_string(self):
        subslices = []
        slice_start = self.start

        for escape_char_position in self.escape_char_positions:
            subslices.append(self.string[slice_start:escape_char_position])
            slice_start = escape_char_position + 1

        subslices.append(self.string[slice_start : self.pos])
        self.escape_char_positions = []

        return "".join(subslices)

    def next(self):
        if self.pos >= self.length:
            return self.EOS

        char = self.string[self.pos]
        self.pos += 1
        return char

    def emit(self, type_):
        self.lexemes.append(
            {
                "type": type_,
                "string": self.slice_string(),
                "start": self.start,
                "end": self.pos,
            }
        )
        self.start = self.pos

    def escape_character(self):
        self.escape_char_positions.append(self.pos - 1)
        self.pos += 1

    def lex_field(self):
        self.backup()
        self.emit(self.FIELD)
        self.ignore()
        return self.lex_text

    def lex_term(self):
        if self.width > 1:
            self.backup()
            self.emit(self.TERM)

        self.ignore()

        return self.lex_text

    def lex_edit_distance(self):
        self.ignore()
        self.accept_digit_run()
        self.emit(self.EDIT_DISTANCE)
        return self.lex_text

    def lex_boost(self):
        self.ignore()
        self.accept_digit_run()
        self.emit(self.BOOST)
        return self.lex_text

    def lex_EOS(self):
        if self.width > 0:
            self.emit(self.TERM)

    def lex_text(self):
        while True:
            char = self.next()
            if char == self.EOS:
                return self.lex_EOS

            if ord(char) == 92:  # Escape character is '\'
                self.escape_character()
                continue

            if char == ":":
                return self.lex_field

            if char == "~":
                self.backup()
                if self.width > 0:
                    self.emit(self.TERM)

                return self.lex_edit_distance

            if char == "^":
                self.backup()
                if self.width > 0:
                    self.emit(self.TERM)

                return self.lex_boost

            # '+' indicates term presence is required, check for length to
            # ensure only a leading '+' is considered
            if char == "+" and self.width == 1:
                self.emit(self.PRESENCE)
                return self.lex_text

            # '-' indicates term presence is prohibited
            if char == "-" and self.width == 1:
                self.emit(self.PRESENCE)
                return self.lex_text

            if default_separator(char):
                return self.lex_term



from lunr.pipeline import Pipeline

WORDS = {
    "a",
    "able",
    "about",
    "across",
    "after",
    "all",
    "almost",
    "also",
    "am",
    "among",
    "an",
    "and",
    "any",
    "are",
    "as",
    "at",
    "be",
    "because",
    "been",
    "but",
    "by",
    "can",
    "cannot",
    "could",
    "dear",
    "did",
    "do",
    "does",
    "either",
    "else",
    "ever",
    "every",
    "for",
    "from",
    "get",
    "got",
    "had",
    "has",
    "have",
    "he",
    "her",
    "hers",
    "him",
    "his",
    "how",
    "however",
    "i",
    "if",
    "in",
    "into",
    "is",
    "it",
    "its",
    "just",
    "least",
    "let",
    "like",
    "likely",
    "may",
    "me",
    "might",
    "most",
    "must",
    "my",
    "neither",
    "no",
    "nor",
    "not",
    "of",
    "off",
    "often",
    "on",
    "only",
    "or",
    "other",
    "our",
    "own",
    "rather",
    "said",
    "say",
    "says",
    "she",
    "should",
    "since",
    "so",
    "some",
    "than",
    "that",
    "the",
    "their",
    "them",
    "then",
    "there",
    "these",
    "they",
    "this",
    "tis",
    "to",
    "too",
    "twas",
    "us",
    "wants",
    "was",
    "we",
    "were",
    "what",
    "when",
    "where",
    "which",
    "while",
    "who",
    "whom",
    "why",
    "will",
    "with",
    "would",
    "yet",
    "you",
    "your",
}


def generate_stop_word_filter(stop_words, language=None):
    """Builds a stopWordFilter function from the provided list of stop words.

    The built in `stop_word_filter` is built using this factory and can be used
    to generate custom `stop_word_filter` for applications or non English
    languages.
    """

    def stop_word_filter(token, i=None, tokens=None):
        if token and str(token) not in stop_words:
            return token

    # camelCased for for compatibility with lunr.js
    label = (
        "stopWordFilter-{}".format(language)
        if language is not None
        else "stopWordFilter"
    )
    Pipeline.register_function(stop_word_filter, label)
    return stop_word_filter


stop_word_filter = generate_stop_word_filter(WORDS)



import pytest

from lunr.query import Query, QueryPresence
from lunr.query_parser import QueryParser
from lunr.exceptions import QueryParseError


def parse(q):
    query = Query(["title", "body"])
    parser = QueryParser(q, query)

    parser.parse()
    return query.clauses


class TestQueryParser:
    def test_parse_empty_string(self):
        clauses = parse("")
        assert len(clauses) == 0

    def test_parse_single_term(self):
        clauses = parse("foo")
        assert len(clauses) == 1
        clause = clauses[0]
        assert clause.term == "foo"
        assert clause.fields == ["title", "body"]
        assert clause.use_pipeline is True
        assert clause.presence is QueryPresence.OPTIONAL

    def test_parse_single_term_uppercase(self):
        clauses = parse("FOO")
        assert len(clauses) == 1
        clause = clauses[0]
        assert clause.term == "foo"
        assert clause.fields == ["title", "body"]
        assert clause.use_pipeline is True

    def test_parse_single_term_with_wildcard(self):
        clauses = parse("fo*")
        assert len(clauses) == 1
        clause = clauses[0]
        assert clause.term == "fo*"
        assert clause.use_pipeline is False

    def test_multiple_terms(self):
        clauses = parse("foo bar")
        assert len(clauses) == 2
        assert clauses[0].term == "foo"
        assert clauses[1].term == "bar"

    def test_term_with_presence_required_adds_required_clause(self):
        clauses = parse("+foo")
        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].boost == 1
        assert clauses[0].fields == ["title", "body"]
        assert clauses[0].presence == QueryPresence.REQUIRED

    def test_term_with_presence_required_adds_prohibited_clause(self):
        clauses = parse("-foo")
        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].boost == 1
        assert clauses[0].fields == ["title", "body"]
        assert clauses[0].presence == QueryPresence.PROHIBITED

    def test_term_scoped_by_field_with_presence_required(self):
        clauses = parse("+title:foo")
        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].boost == 1
        assert clauses[0].fields == ["title"]
        assert clauses[0].presence == QueryPresence.REQUIRED

    def test_term_scoped_by_field_with_presence_prohibited(self):
        clauses = parse("-title:foo")
        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].boost == 1
        assert clauses[0].fields == ["title"]
        assert clauses[0].presence == QueryPresence.PROHIBITED

    def test_multiple_terms_with_presence_creates_two_clauses(self):
        clauses = parse("+foo +bar")
        assert len(clauses) == 2
        assert clauses[0].term == "foo"
        assert clauses[1].term == "bar"
        assert clauses[0].presence == QueryPresence.REQUIRED
        assert clauses[1].presence == QueryPresence.REQUIRED

    def test_unknown_field(self):
        with pytest.raises(QueryParseError):
            parse("unknown:foo")

    def test_field_without_a_term(self):
        with pytest.raises(QueryParseError):
            parse("title:")

    def test_field_twice(self):
        with pytest.raises(QueryParseError):
            parse("title:title:")

    def test_term_with_field(self):
        clauses = parse("title:foo")
        assert len(clauses) == 1
        assert clauses[0].fields == ["title"]

    def test_uppercase_field_with_uppercase_term(self):
        query = Query(["TITLE"])
        parser = QueryParser("TITLE:FOO", query)

        parser.parse()
        clauses = query.clauses

        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].fields == ["TITLE"]

    def test_multiple_terms_scoped_to_different_fields(self):
        clauses = parse("title:foo body:bar")

        assert len(clauses) == 2
        assert clauses[0].fields == ["title"]
        assert clauses[1].fields == ["body"]

        assert clauses[0].term == "foo"
        assert clauses[1].term == "bar"

    def test_single_term_with_edit_distance(self):
        clauses = parse("foo~2")

        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].fields == ["title", "body"]
        assert clauses[0].edit_distance == 2

    def test_multiple_terms_with_edit_distance(self):
        clauses = parse("foo~2 bar~3")

        assert len(clauses) == 2
        assert clauses[0].fields == ["title", "body"]
        assert clauses[1].fields == ["title", "body"]

        assert clauses[0].term == "foo"
        assert clauses[1].term == "bar"

        assert clauses[0].edit_distance == 2
        assert clauses[1].edit_distance == 3

    def test_single_term_scoped_to_field_with_edit_distance(self):
        clauses = parse("title:foo~2")

        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].fields == ["title"]
        assert clauses[0].edit_distance == 2

    def test_non_numeric_edit_distance(self):
        with pytest.raises(QueryParseError):
            parse("foo~a")

    def test_edit_distance_without_a_term(self):
        with pytest.raises(QueryParseError):
            parse("~2")

    def test_single_term_with_boost(self):
        clauses = parse("foo^2")

        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].fields == ["title", "body"]
        assert clauses[0].boost == 2

    def test_non_numeric_boost(self):
        with pytest.raises(QueryParseError):
            parse("foo^a")

    def test_boost_without_a_term(self):
        with pytest.raises(QueryParseError):
            parse("^2")

    def test_multiple_terms_with_boost(self):
        clauses = parse("foo^2 bar^3")

        assert len(clauses) == 2
        assert clauses[0].fields == ["title", "body"]
        assert clauses[1].fields == ["title", "body"]

        assert clauses[0].term == "foo"
        assert clauses[1].term == "bar"

        assert clauses[0].boost == 2
        assert clauses[1].boost == 3

    def test_term_scoped_by_field_with_boost(self):
        clauses = parse("title:foo^2")

        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].fields == ["title"]
        assert clauses[0].boost == 2

    def test_term_with_boost_and_edit_distance(self):
        clauses = parse("foo^2~3")

        assert len(clauses) == 1
        assert clauses[0].term == "foo"
        assert clauses[0].fields == ["title", "body"]
        assert clauses[0].edit_distance == 3
        assert clauses[0].boost == 2

    def test_edit_distance_followed_by_presence(self):
        clauses = parse("foo~10 +bar")

        assert len(clauses) == 2

        assert clauses[0].fields == ["title", "body"]
        assert clauses[1].fields == ["title", "body"]

        assert clauses[0].term == "foo"
        assert clauses[1].term == "bar"

        assert clauses[0].edit_distance == 10
        assert clauses[1].edit_distance == 0

        assert clauses[0].presence == QueryPresence.OPTIONAL
        assert clauses[1].presence == QueryPresence.REQUIRED

    def test_boost_followed_by_presence(self):
        clauses = parse("foo^10 +bar")

        assert len(clauses) == 2

        assert clauses[0].fields == ["title", "body"]
        assert clauses[1].fields == ["title", "body"]

        assert clauses[0].term == "foo"
        assert clauses[1].term == "bar"

        assert clauses[0].boost == 10
        assert clauses[1].boost == 1

        assert clauses[0].presence == QueryPresence.OPTIONAL
        assert clauses[1].presence == QueryPresence.REQUIRED
